@@ -1,14 +1,17 @@
+# SPDX-License-Identifier: GPL-2.0-only
 # This file is part of Scapy
-# See http://www.secdev.org/projects/scapy for more information
+# See https://scapy.net/ for more information
 # Copyright (C) Philippe Biondi <phil@secdev.org>
-# This program is published under a GPLv2 license
 
 """
 DHCP (Dynamic Host Configuration Protocol) and BOOTP
+
+Implements:
+- rfc951 - BOOTSTRAP PROTOCOL (BOOTP)
+- rfc1542 - Clarifications and Extensions for the Bootstrap Protocol
+- rfc1533 - DHCP Options and BOOTP Vendor Extensions
 """
 
-from __future__ import absolute_import
-from __future__ import print_function
 try:
     from collections.abc import Iterable
 except ImportError:
@@ -17,44 +20,76 @@ except ImportError:
 import random
 import struct
 
+import socket
+import re
+
 from scapy.ansmachine import AnsweringMachine
 from scapy.base_classes import Net
 from scapy.compat import chb, orb, bytes_encode
-from scapy.fields import ByteEnumField, ByteField, Field, FieldListField, \
-    FlagsField, IntField, IPField, ShortField, StrField
+from scapy.fields import (
+    ByteEnumField,
+    ByteField,
+    Field,
+    FieldListField,
+    FlagsField,
+    IntField,
+    IPField,
+    ShortField,
+    StrEnumField,
+    StrField,
+    StrFixedLenField,
+    XIntField,
+)
 from scapy.layers.inet import UDP, IP
-from scapy.layers.l2 import Ether
+from scapy.layers.l2 import Ether, HARDWARE_TYPES
 from scapy.packet import bind_layers, bind_bottom_up, Packet
-from scapy.utils import atol, itom, ltoa, sane
-from scapy.volatile import RandBin, RandField, RandNum, RandNumExpo
+from scapy.utils import atol, itom, ltoa, sane, str2mac, mac2str
+from scapy.volatile import (
+    RandBin,
+    RandByte,
+    RandField,
+    RandIP,
+    RandInt,
+    RandNum,
+    RandNumExpo,
+)
 
-from scapy.arch import get_if_raw_hwaddr
-from scapy.sendrecv import srp1, sendp
+from scapy.arch import get_if_hwaddr
+from scapy.sendrecv import srp1
 from scapy.error import warning
-import scapy.modules.six as six
-from scapy.modules.six.moves import range
 from scapy.config import conf
 
 dhcpmagic = b"c\x82Sc"
 
 
+class _BOOTP_chaddr(StrFixedLenField):
+    def i2repr(self, pkt, v):
+        if pkt.htype == 1:  # Ethernet
+            if v[6:] == b"\x00" * 10:  # Default padding
+                return "%s (+ 10 nul pad)" % str2mac(v[:6])
+            else:
+                return "%s (pad: %s)" % (str2mac(v[:6]), v[6:])
+        return super(_BOOTP_chaddr, self).i2repr(pkt, v)
+
+
 class BOOTP(Packet):
     name = "BOOTP"
-    fields_desc = [ByteEnumField("op", 1, {1: "BOOTREQUEST", 2: "BOOTREPLY"}),
-                   ByteField("htype", 1),
-                   ByteField("hlen", 6),
-                   ByteField("hops", 0),
-                   IntField("xid", 0),
-                   ShortField("secs", 0),
-                   FlagsField("flags", 0, 16, "???????????????B"),
-                   IPField("ciaddr", "0.0.0.0"),
-                   IPField("yiaddr", "0.0.0.0"),
-                   IPField("siaddr", "0.0.0.0"),
-                   IPField("giaddr", "0.0.0.0"),
-                   Field("chaddr", b"", "16s"),
-                   Field("sname", b"", "64s"),
-                   Field("file", b"", "128s"),
-                   StrField("options", b"")]
+    fields_desc = [
+        ByteEnumField("op", 1, {1: "BOOTREQUEST", 2: "BOOTREPLY"}),
+        ByteEnumField("htype", 1, HARDWARE_TYPES),
+        ByteField("hlen", 6),
+        ByteField("hops", 0),
+        XIntField("xid", 0),
+        ShortField("secs", 0),
+        FlagsField("flags", 0, 16, "???????????????B"),
+        IPField("ciaddr", "0.0.0.0"),
+        IPField("yiaddr", "0.0.0.0"),
+        IPField("siaddr", "0.0.0.0"),
+        IPField("giaddr", "0.0.0.0"),
+        _BOOTP_chaddr("chaddr", b"", length=16),
+        StrFixedLenField("sname", b"", length=64),
+        StrFixedLenField("file", b"", length=128),
+        StrEnumField("options", b"", {dhcpmagic: "DHCP magic"})]
 
     def guess_payload_class(self, payload):
         if self.options[:len(dhcpmagic)] == dhcpmagic:
@@ -81,17 +116,94 @@ class BOOTP(Packet):
 
 
 class _DHCPParamReqFieldListField(FieldListField):
+    def randval(self):
+        class _RandReqFieldList(RandField):
+            def _fix(self):
+                return [RandByte()] * int(RandByte())
+        return _RandReqFieldList()
+
+
+class RandClasslessStaticRoutesField(RandField):
+    """
+    A RandValue for classless static routes
+    """
+
+    def _fix(self):
+        return "%s/%d:%s" % (RandIP(), RandNum(0, 32), RandIP())
+
+
+class ClasslessFieldListField(FieldListField):
+    def randval(self):
+        class _RandClasslessField(RandField):
+            def _fix(self):
+                return [RandClasslessStaticRoutesField()] * int(RandNum(1, 28))
+        return _RandClasslessField()
+
+
+class ClasslessStaticRoutesField(Field):
+    """
+    RFC 3442 defines classless static routes as up to 9 bytes per entry:
+
+    # Code Len Destination 1    Router 1
+    +-----+---+----+-----+----+----+----+----+----+
+    | 121 | n | d1 | ... | dN | r1 | r2 | r3 | r4 |
+    +-----+---+----+-----+----+----+----+----+----+
+
+    Destination first byte contains one octet describing the width followed
+    by all the significant octets of the subnet.
+    """
+
+    def m2i(self, pkt, x):
+        # type: (Packet, bytes) -> str
+        # b'\x20\x01\x02\x03\x04\t\x08\x07\x06' -> (1.2.3.4/32:9.8.7.6)
+        prefix = orb(x[0])
+
+        octets = (prefix + 7) // 8
+        # Create the destination IP by using the number of octets
+        # and padding up to 4 bytes to ensure a valid IP.
+        dest = x[1:1 + octets]
+        dest = socket.inet_ntoa(dest.ljust(4, b'\x00'))
+
+        router = x[1 + octets:5 + octets]
+        router = socket.inet_ntoa(router)
+
+        return dest + "/" + str(prefix) + ":" + router
+
+    def i2m(self, pkt, x):
+        # type: (Packet, str) -> bytes
+        # (1.2.3.4/32:9.8.7.6) -> b'\x20\x01\x02\x03\x04\t\x08\x07\x06'
+        if not x:
+            return b''
+
+        spx = re.split('/|:', str(x))
+        prefix = int(spx[1])
+        # if prefix is invalid value ( 0 > prefix > 32 ) then break
+        if prefix > 32 or prefix < 0:
+            warning("Invalid prefix value: %d (0x%x)", prefix, prefix)
+            return b''
+        octets = (prefix + 7) // 8
+        dest = socket.inet_aton(spx[0])[:octets]
+        router = socket.inet_aton(spx[2])
+        return struct.pack('b', prefix) + dest + router
+
     def getfield(self, pkt, s):
-        ret = []
-        while s:
-            s, val = FieldListField.getfield(self, pkt, s)
-            ret.append(val)
-        return b"", [x[0] for x in ret]
+        prefix = orb(s[0])
+        route_len = 5 + (prefix + 7) // 8
+        return s[route_len:], self.m2i(pkt, s[:route_len])
+
+    def addfield(self, pkt, s, val):
+        return s + self.i2m(pkt, val)
+
+    def randval(self):
+        return RandClasslessStaticRoutesField()
+
 
 # DHCP_UNKNOWN, DHCP_IP, DHCP_IPLIST, DHCP_TYPE \
 # = range(4)
 #
 
+# DHCP Options and BOOTP Vendor Extensions
+
 
 DHCPTypes = {
     1: "discover",
@@ -156,7 +268,7 @@ DHCPOptions = {
     43: "vendor_specific",
     44: IPField("NetBIOS_server", "0.0.0.0"),
     45: IPField("NetBIOS_dist_server", "0.0.0.0"),
-    46: ByteField("static-routes", 100),
+    46: ByteField("NetBIOS_node_type", 100),
     47: "netbios-scope",
     48: IPField("font-servers", "0.0.0.0"),
     49: IPField("x-display-manager", "0.0.0.0"),
@@ -165,7 +277,9 @@ DHCPOptions = {
     52: ByteField("dhcp-option-overload", 100),
     53: ByteEnumField("message-type", 1, DHCPTypes),
     54: IPField("server_id", "0.0.0.0"),
-    55: _DHCPParamReqFieldListField("param_req_list", [], ByteField("opcode", 0), length_from=lambda x: 1),  # noqa: E501
+    55: _DHCPParamReqFieldListField(
+        "param_req_list", [],
+        ByteField("opcode", 0)),
     56: "error_message",
     57: ShortField("max_dhcp_size", 1500),
     58: IntField("renewal_time", 21600),
@@ -175,6 +289,7 @@ DHCPOptions = {
     62: "nwip-domain-name",
     64: "NISplus_domain",
     65: IPField("NISplus_server", "0.0.0.0"),
+    66: "tftp_server_name",
     67: StrField("boot-file-name", ""),
     68: IPField("mobile-ip-home-agent", "0.0.0.0"),
     69: IPField("SMTP_server", "0.0.0.0"),
@@ -185,8 +300,10 @@ DHCPOptions = {
     74: IPField("IRC_server", "0.0.0.0"),
     75: IPField("StreetTalk_server", "0.0.0.0"),
     76: IPField("StreetTalk_Dir_Assistance", "0.0.0.0"),
+    77: "user_class",
     78: "slp_service_agent",
     79: "slp_service_scope",
+    80: "rapid_commit",
     81: "client_FQDN",
     82: "relay_agent_information",
     85: IPField("nds-server", "0.0.0.0"),
@@ -202,21 +319,29 @@ DHCPOptions = {
     98: StrField("uap-servers", ""),
     100: StrField("pcode", ""),
     101: StrField("tcode", ""),
+    108: IntField("ipv6-only-preferred", 0),
     112: IPField("netinfo-server-address", "0.0.0.0"),
     113: StrField("netinfo-server-tag", ""),
-    114: StrField("default-url", ""),
+    114: StrField("captive-portal", ""),
     116: ByteField("auto-config", 0),
     117: ShortField("name-service-search", 0,),
     118: IPField("subnet-selection", "0.0.0.0"),
+    121: ClasslessFieldListField(
+        "classless_static_routes",
+        [],
+        ClasslessStaticRoutesField("route", 0)),
     124: "vendor_class",
     125: "vendor_specific_information",
+    128: IPField("tftp_server_ip_address", "0.0.0.0"),
     136: IPField("pana-agent", "0.0.0.0"),
     137: "v4-lost",
     138: IPField("capwap-ac-v4", "0.0.0.0"),
     141: "sip_ua_service_domains",
     146: "rdnss-selection",
+    150: IPField("tftp_server_address", "0.0.0.0"),
     159: "v4-portparams",
     160: StrField("v4-captive-portal", ""),
+    161: StrField("mud-url", ""),
     208: "pxelinux_magic",
     209: "pxelinux_configuration_file",
     210: "pxelinux_path_prefix",
@@ -228,16 +353,16 @@ DHCPOptions = {
 
 DHCPRevOptions = {}
 
-for k, v in six.iteritems(DHCPOptions):
+for k, v in DHCPOptions.items():
     if isinstance(v, str):
         n = v
         v = None
     else:
         n = v.name
     DHCPRevOptions[n] = (k, v)
-del(n)
-del(v)
-del(k)
+del n
+del v
+del k
 
 
 class RandDHCPOptions(RandField):
@@ -248,7 +373,7 @@ class RandDHCPOptions(RandField):
         if rndstr is None:
             rndstr = RandBin(RandNum(0, 255))
         self.rndstr = rndstr
-        self._opts = list(six.itervalues(DHCPOptions))
+        self._opts = list(DHCPOptions.values())
         self._opts.remove("pad")
         self._opts.remove("end")
 
@@ -259,11 +384,20 @@ class RandDHCPOptions(RandField):
             if isinstance(o, str):
                 op.append((o, self.rndstr * 1))
             else:
-                op.append((o.name, o.randval()._fix()))
+                r = o.randval()._fix()
+                if isinstance(r, bytes):
+                    r = r[:255]
+                op.append((o.name, r))
         return op
 
 
 class DHCPOptionsField(StrField):
+    """
+    A field that builds and dissects DHCP options.
+    The internal value is a list of tuples with the format
+    [("option_name", <option_value>), ...]
+    Where expected names and values can be found using `DHCPOptions`
+    """
     islist = 1
 
     def i2repr(self, pkt, x):
@@ -275,8 +409,7 @@ class DHCPOptionsField(StrField):
                     vv = ",".join(f.i2repr(pkt, val) for val in v[1:])
                 else:
                     vv = ",".join(repr(val) for val in v[1:])
-                r = "%s=%s" % (v[0], vv)
-                s.append(r)
+                s.append("%s=%s" % (v[0], vv))
             else:
                 s.append(sane(v))
         return "[%s]" % (" ".join(s))
@@ -309,6 +442,16 @@ class DHCPOptionsField(StrField):
                 else:
                     olen = orb(x[1])
                     lval = [f.name]
+
+                    if olen == 0:
+                        try:
+                            _, val = f.getfield(pkt, b'')
+                        except Exception:
+                            opt.append(x)
+                            break
+                        else:
+                            lval.append(val)
+
                     try:
                         left = x[2:olen + 2]
                         while left:
@@ -349,8 +492,7 @@ class DHCPOptionsField(StrField):
                     warning("Unknown field option %s", name)
                     continue
 
-                s += chb(onum)
-                s += chb(len(oval))
+                s += struct.pack("!BB", onum, len(oval))
                 s += oval
 
             elif (isinstance(o, str) and o in DHCPRevOptions and
@@ -364,11 +506,20 @@ class DHCPOptionsField(StrField):
                 warning("Malformed option %s", o)
         return s
 
+    def randval(self):
+        return RandDHCPOptions()
+
 
 class DHCP(Packet):
     name = "DHCP options"
     fields_desc = [DHCPOptionsField("options", b"")]
 
+    def mysummary(self):
+        for id in self.options:
+            if isinstance(id, tuple) and id[0] == "message-type":
+                return "DHCP %s" % DHCPTypes.get(id[1], "").capitalize()
+        return super(DHCP, self).mysummary()
+
 
 bind_layers(UDP, BOOTP, dport=67, sport=68)
 bind_layers(UDP, BOOTP, dport=68, sport=67)
@@ -377,26 +528,89 @@ bind_layers(BOOTP, DHCP, options=b'c\x82Sc')
 
 
 @conf.commands.register
-def dhcp_request(iface=None, **kargs):
-    """Send a DHCP discover request and return the answer"""
+def dhcp_request(hw=None,
+                 req_type='discover',
+                 server_id=None,
+                 requested_addr=None,
+                 hostname=None,
+                 iface=None,
+                 **kargs):
+    """
+    Send a DHCP discover request and return the answer.
+
+    Usage::
+
+        >>> dhcp_request()  # send DHCP discover
+        >>> dhcp_request(req_type='request',
+        ...              requested_addr='10.53.4.34')  # send DHCP request
+    """
     if conf.checkIPaddr:
         warning(
             "conf.checkIPaddr is enabled, may not be able to match the answer"
         )
-    if iface is None:
-        iface = conf.iface
-    fam, hw = get_if_raw_hwaddr(iface)
-    return srp1(Ether(dst="ff:ff:ff:ff:ff:ff") / IP(src="0.0.0.0", dst="255.255.255.255") / UDP(sport=68, dport=67) /  # noqa: E501
-                BOOTP(chaddr=hw) / DHCP(options=[("message-type", "discover"), "end"]), iface=iface, **kargs)  # noqa: E501
+    if hw is None:
+        if iface is None:
+            iface = conf.iface
+        hw = get_if_hwaddr(iface)
+    dhcp_options = [
+        ('message-type', req_type),
+        ('client_id', b'\x01' + mac2str(hw)),
+    ]
+    if requested_addr is not None:
+        dhcp_options.append(('requested_addr', requested_addr))
+    elif req_type == 'request':
+        warning("DHCP Request without requested_addr will likely be ignored")
+    if server_id is not None:
+        dhcp_options.append(('server_id', server_id))
+    if hostname is not None:
+        dhcp_options.extend([
+            ('hostname', hostname),
+            ('client_FQDN', b'\x00\x00\x00' + bytes_encode(hostname)),
+        ])
+    dhcp_options.extend([
+        ('vendor_class_id', b'MSFT 5.0'),
+        ('param_req_list', [
+            1, 3, 6, 15, 31, 33, 43, 44, 46, 47, 119, 121, 249, 252
+        ]),
+        'end'
+    ])
+    return srp1(
+        Ether(dst="ff:ff:ff:ff:ff:ff", src=hw) /
+        IP(src="0.0.0.0", dst="255.255.255.255") /
+        UDP(sport=68, dport=67) /
+        BOOTP(chaddr=hw, xid=RandInt(), flags="B") /
+        DHCP(options=dhcp_options),
+        iface=iface, **kargs
+    )
 
 
 class BOOTP_am(AnsweringMachine):
     function_name = "bootpd"
     filter = "udp and port 68 and port 67"
-    send_function = staticmethod(sendp)
 
-    def parse_options(self, pool=Net("192.168.1.128/25"), network="192.168.1.0/24", gw="192.168.1.1",  # noqa: E501
-                      domain="localnet", renewal_time=60, lease_time=1800):
+    def parse_options(self,
+                      pool=Net("192.168.1.128/25"),
+                      network="192.168.1.0/24",
+                      gw="192.168.1.1",
+                      nameserver=None,
+                      domain=None,
+                      renewal_time=60,
+                      lease_time=1800,
+                      **kwargs):
+        """
+        :param pool: the range of addresses to distribute. Can be a Net,
+                     a list of IPs or a string (always gives the same IP).
+        :param network: the subnet range
+        :param gw: the gateway IP (can be None)
+        :param nameserver: the DNS server IP (by default, same than gw)
+        :param domain: the domain to advertise (can be None)
+
+        Other DHCP parameters can be passed as kwargs. See DHCPOptions in dhcp.py.
+        For instance::
+
+            dhcpd(pool=Net("10.0.10.0/24"), network="10.0.0.0/8", gw="10.0.10.1",
+                  classless_static_routes=["1.2.3.4/32:9.8.7.6"])
+        """
         self.domain = domain
         netw, msk = (network.split("/") + ["32"])[:2]
         msk = itom(int(msk))
@@ -404,10 +618,11 @@ class BOOTP_am(AnsweringMachine):
         self.network = ltoa(atol(netw) & msk)
         self.broadcast = ltoa(atol(self.network) | (0xffffffff & ~msk))
         self.gw = gw
-        if isinstance(pool, six.string_types):
+        self.nameserver = nameserver or gw
+        if isinstance(pool, str):
             pool = Net(pool)
         if isinstance(pool, Iterable):
-            pool = [k for k in pool if k not in [gw, self.network, self.broadcast]]  # noqa: E501
+            pool = [k for k in pool if k not in [gw, self.network, self.broadcast]]
             pool.reverse()
         if len(pool) == 1:
             pool, = pool
@@ -415,6 +630,7 @@ class BOOTP_am(AnsweringMachine):
         self.lease_time = lease_time
         self.renewal_time = renewal_time
         self.leases = {}
+        self.kwargs = kwargs
 
     def is_request(self, req):
         if not req.haslayer(BOOTP):
@@ -424,7 +640,7 @@ class BOOTP_am(AnsweringMachine):
             return 0
         return 1
 
-    def print_reply(self, req, reply):
+    def print_reply(self, _, reply):
         print("Reply %s to %s" % (reply.getlayer(IP).dst, reply.dst))
 
     def make_reply(self, req):
@@ -442,7 +658,7 @@ class BOOTP_am(AnsweringMachine):
         repb.siaddr = self.gw
         repb.ciaddr = self.gw
         repb.giaddr = self.gw
-        del(repb.payload)
+        del repb.payload
         rep = Ether(dst=mac) / IP(dst=ip) / UDP(sport=req.dport, dport=req.sport) / repb  # noqa: E501
         return rep
 
@@ -453,18 +669,26 @@ class DHCP_am(BOOTP_am):
     def make_reply(self, req):
         resp = BOOTP_am.make_reply(self, req)
         if DHCP in req:
-            dhcp_options = [(op[0], {1: 2, 3: 5}.get(op[1], op[1]))
-                            for op in req[DHCP].options
-                            if isinstance(op, tuple) and op[0] == "message-type"]  # noqa: E501
-            dhcp_options += [("server_id", self.gw),
-                             ("domain", self.domain),
-                             ("router", self.gw),
-                             ("name_server", self.gw),
-                             ("broadcast_address", self.broadcast),
-                             ("subnet_mask", self.netmask),
-                             ("renewal_time", self.renewal_time),
-                             ("lease_time", self.lease_time),
-                             "end"
-                             ]
+            dhcp_options = [
+                (op[0], {1: 2, 3: 5}.get(op[1], op[1]))
+                for op in req[DHCP].options
+                if isinstance(op, tuple) and op[0] == "message-type"
+            ]
+            dhcp_options += [
+                x for x in [
+                    ("server_id", self.gw),
+                    ("domain", self.domain),
+                    ("router", self.gw),
+                    ("name_server", self.nameserver),
+                    ("broadcast_address", self.broadcast),
+                    ("subnet_mask", self.netmask),
+                    ("renewal_time", self.renewal_time),
+                    ("lease_time", self.lease_time),
+                ]
+                if x[1] is not None
+            ]
+            if self.kwargs:
+                dhcp_options += self.kwargs.items()
+            dhcp_options.append("end")
             resp /= DHCP(options=dhcp_options)
         return resp
